from __future__ import annotations

import subprocess
import sys
import textwrap
from pathlib import Path

import pytest

from cook.cli import main
from cook.context import Context


@pytest.fixture
def project(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> Path:
    """Set up a temp project directory and chdir into it."""
    monkeypatch.chdir(tmp_path)
    return tmp_path


def _write_recipe(project: Path, code: str, name: str = "recipe.py") -> Path:
    recipe = project / name
    recipe.write_text(textwrap.dedent(code))
    return recipe


def test_exec_basic(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    outfile = project / "out.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="hello", cmd="echo hello > {outfile}", outputs=["{outfile}"])
        """,
    )
    rc = main(["run", "hello"])
    assert rc == 0
    assert outfile.exists()
    assert outfile.read_text().strip() == "hello"


def test_exec_pattern_matching(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    out_a = project / "a.txt"
    out_b = project / "b.txt"
    out_other = project / "other.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="compile-a", cmd="echo a > {out_a}", outputs=["{out_a}"])
        ctx.sh(name="compile-b", cmd="echo b > {out_b}", outputs=["{out_b}"])
        ctx.sh(name="test-all", cmd="echo other > {out_other}", outputs=["{out_other}"])
        """,
    )
    rc = main(["run", "compile-*"])
    assert rc == 0
    assert out_a.exists()
    assert out_b.exists()
    assert not out_other.exists()


def test_exec_multiple_patterns(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    out_a = project / "a.txt"
    out_b = project / "b.txt"
    out_c = project / "c.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="compile-a", cmd="echo a > {out_a}", outputs=["{out_a}"])
        ctx.sh(name="test-b", cmd="echo b > {out_b}", outputs=["{out_b}"])
        ctx.sh(name="lint-c", cmd="echo c > {out_c}", outputs=["{out_c}"])
        """,
    )
    rc = main(["run", "compile-*", "test-*"])
    assert rc == 0
    assert out_a.exists()
    assert out_b.exists()
    assert not out_c.exists()


def test_exec_regex_pattern(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    out_a = project / "a.txt"
    out_b = project / "b.txt"
    out_other = project / "other.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="compile-a", cmd="echo a > {out_a}", outputs=["{out_a}"])
        ctx.sh(name="compile-b", cmd="echo b > {out_b}", outputs=["{out_b}"])
        ctx.sh(name="test-all", cmd="echo other > {out_other}", outputs=["{out_other}"])
        """,
    )
    rc = main(["run", "--re", "^compile-"])
    assert rc == 0
    assert out_a.exists()
    assert out_b.exists()
    assert not out_other.exists()


def test_exec_regex_invalid(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="t", cmd="true", outputs=[])
        """,
    )
    rc = main(["run", "--re", "[invalid"])
    assert rc == 1
    assert "Invalid regex" in capsys.readouterr().err


def test_exec_with_default_config(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    outfile = project / "out.txt"
    (project / "cook.toml").write_text('[cook]\ndefault = "hello"\n')
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="hello", cmd="echo hi > {outfile}", outputs=["{outfile}"])
        """,
    )
    rc = main(["run"])
    assert rc == 0
    assert outfile.exists()


def test_exec_no_pattern_no_default(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="task1", cmd="true", outputs=[])
        """,
    )
    rc = main(["run"])
    assert rc == 1
    assert "No target pattern" in capsys.readouterr().err


def test_exec_no_matches(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="task1", cmd="true", outputs=[])
        """,
    )
    rc = main(["run", "nonexistent-*"])
    assert rc == 1
    assert "matched no tasks" in capsys.readouterr().err


def test_exec_dry_run(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    outfile = project / "out.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="build", cmd="echo x > {outfile}", outputs=["{outfile}"])
        """,
    )
    rc = main(["run", "--dry-run", "build"])
    assert rc == 0
    captured = capsys.readouterr().err
    assert "STALE (would run)" in captured
    # File should NOT be created
    assert not outfile.exists()


def test_exec_dry_run_with_existing_store(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    outfile = project / "out.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="build", cmd="echo x > {outfile}", outputs=["{outfile}"])
        """,
    )
    # First run to populate store
    rc = main(["run", "build"])
    assert rc == 0
    capsys.readouterr()

    # Now dry run should show up-to-date
    rc = main(["run", "--dry-run", "build"])
    assert rc == 0
    captured = capsys.readouterr().err
    assert "up-to-date" in captured


def test_inspect(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    out_a = project / "a.txt"
    out_b = project / "b.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        a = ctx.sh(name="step-a", cmd="echo a > {out_a}", outputs=["{out_a}"])
        ctx.sh(name="step-b", cmd="echo b > {out_b}", inputs=[a], outputs=["{out_b}"])
        """,
    )
    rc = main(["inspect", "*"])
    assert rc == 0
    captured = capsys.readouterr().out
    assert "[step-a] STALE" in captured
    assert "[step-b] STALE" in captured
    assert "deps: step-a" in captured


def test_inspect_shows_never_run(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="fresh-task", cmd="true", outputs=["out.txt"])
        """,
    )
    rc = main(["inspect", "fresh-task"])
    assert rc == 0
    captured = capsys.readouterr().out
    assert "never run" in captured


def test_inspect_shows_why_always_run(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="check", cmd="true")
        """,
    )
    rc = main(["inspect", "check"])
    assert rc == 0
    captured = capsys.readouterr().out
    assert "always-run" in captured


def test_inspect_shows_why_output_missing(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    outfile = project / "out.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="build", cmd="echo x > {outfile}", outputs=["{outfile}"])
        """,
    )
    # Run first, then delete output
    rc = main(["run", "build"])
    assert rc == 0
    outfile.unlink()
    capsys.readouterr()

    rc = main(["inspect", "build"])
    assert rc == 0
    captured = capsys.readouterr().out
    assert "missing" in captured


def test_inspect_shows_why_always_run_with_store(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    """Always-run reason via staleness_reason (store exists)."""
    outfile = project / "out.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="build", cmd="echo x > {outfile}", outputs=["{outfile}"])
        ctx.sh(name="check", cmd="true")
        """,
    )
    # Run build to create the store, then inspect check
    rc = main(["run", "build"])
    assert rc == 0
    capsys.readouterr()

    rc = main(["inspect", "check"])
    assert rc == 0
    captured = capsys.readouterr().out
    assert "always-run" in captured


def test_inspect_shows_never_run_with_store(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    """Never-run reason via staleness_reason (store exists but no record)."""
    outfile = project / "out.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="built", cmd="echo x > {outfile}", outputs=["{outfile}"])
        ctx.sh(name="new-task", cmd="true", outputs=["new.txt"])
        """,
    )
    # Run only built to create the store
    rc = main(["run", "built"])
    assert rc == 0
    capsys.readouterr()

    rc = main(["inspect", "new-task"])
    assert rc == 0
    captured = capsys.readouterr().out
    assert "never run" in captured


def test_inspect_shows_why_dep_stale(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    out_a = project / "a.txt"
    out_b = project / "b.txt"
    infile = project / "src.txt"
    infile.write_text("original")
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        a = ctx.sh(name="step-a", cmd="cat {infile} > {out_a}", inputs=["{infile}"], outputs=["{out_a}"])
        ctx.sh(name="step-b", cmd="cat {out_a} > {out_b}", inputs=[a], outputs=["{out_b}"])
        """,
    )
    rc = main(["run", "*"])
    assert rc == 0
    capsys.readouterr()

    # Modify input to make step-a stale
    infile.write_text("changed")
    rc = main(["inspect", "step-b"])
    assert rc == 0
    captured = capsys.readouterr().out
    assert "dependency" in captured
    assert "step-a" in captured


def test_inspect_shows_why_input_missing(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    infile = project / "src.txt"
    infile.write_text("data")
    outfile = project / "out.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="build", cmd="cat {infile} > {outfile}", inputs=["{infile}"], outputs=["{outfile}"])
        """,
    )
    rc = main(["run", "build"])
    assert rc == 0
    capsys.readouterr()

    # Delete the input file
    infile.unlink()
    rc = main(["inspect", "build"])
    assert rc == 0
    captured = capsys.readouterr().out
    assert "missing" in captured


def test_inspect_shows_why_digest_changed(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    infile = project / "src.txt"
    infile.write_text("original")
    outfile = project / "out.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="build", cmd="cat {infile} > {outfile}", inputs=["{infile}"], outputs=["{outfile}"])
        """,
    )
    # Run first
    rc = main(["run", "build"])
    assert rc == 0
    capsys.readouterr()

    # Modify input
    infile.write_text("changed")
    rc = main(["inspect", "build"])
    assert rc == 0
    captured = capsys.readouterr().out
    assert "digest changed" in captured


def test_inspect_with_store(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    outfile = project / "out.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="mytask", cmd="echo x > {outfile}", outputs=["{outfile}"])
        """,
    )
    # Run first to populate store
    rc = main(["run", "mytask"])
    assert rc == 0
    capsys.readouterr()

    # Inspect should show up-to-date
    rc = main(["inspect", "mytask"])
    assert rc == 0
    captured = capsys.readouterr().out
    assert "up-to-date" in captured


def test_format_relative_time() -> None:
    from datetime import datetime, timedelta, timezone

    from cook.cli.util import format_relative_time

    now = datetime.now(timezone.utc)
    assert "s ago" in format_relative_time(now - timedelta(seconds=30))
    assert "m ago" in format_relative_time(now - timedelta(minutes=5))
    assert "h ago" in format_relative_time(now - timedelta(hours=3))
    assert "d ago" in format_relative_time(now - timedelta(days=2))


def test_inspect_shows_details(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    """Inspect shows inputs, outputs, command, and execution history."""
    infile = project / "src.txt"
    infile.write_text("data")
    outfile = project / "out.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(
            name="build",
            cmd="cat {infile} > {outfile}",
            inputs=["{infile}"],
            outputs=["{outfile}"],
        )
        """,
    )
    # Run to populate store
    rc = main(["run", "build"])
    assert rc == 0
    capsys.readouterr()

    rc = main(["inspect", "build"])
    assert rc == 0
    captured = capsys.readouterr().out
    assert "[build] up-to-date" in captured
    assert "inputs:" in captured
    assert "outputs:" in captured
    assert "cmd:" in captured
    assert "last started:" in captured
    assert "last succeeded:" in captured


def test_inspect_shows_failure_history(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    """Inspect shows last failed and error message."""
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="bad", cmd="echo oops >&2; exit 1", outputs=["nope.txt"])
        """,
    )
    # Run and fail
    rc = main(["run", "bad"])
    assert rc == 1
    capsys.readouterr()

    rc = main(["inspect", "bad"])
    assert rc == 0
    captured = capsys.readouterr().out
    assert "[bad] STALE" in captured
    assert "last failed:" in captured
    assert "error:" in captured


def test_invalidate(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    outfile = project / "out.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="mytask", cmd="echo hello > {outfile}", outputs=["{outfile}"])
        """,
    )
    # Run first
    rc = main(["run", "mytask"])
    assert rc == 0
    capsys.readouterr()

    # Invalidate
    rc = main(["invalidate", "mytask"])
    assert rc == 0
    captured = capsys.readouterr().err
    assert "Invalidated [mytask]" in captured

    # Run again — should re-execute (not up-to-date)
    rc = main(["run", "mytask"])
    assert rc == 0
    captured = capsys.readouterr().err
    assert "Cooked" in captured


def test_invalidate_no_store(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    """Invalidate without .cook.db should succeed with a message."""
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="t", cmd="true")
        """,
    )
    rc = main(["invalidate", "t"])
    assert rc == 0
    assert "nothing to invalidate" in capsys.readouterr().err.lower()


def test_keep_going(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    good_out = project / "good.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="fail-task", cmd="exit 1", outputs=["{project / "nope.txt"}"])
        ctx.sh(name="good-task", cmd="echo ok > {good_out}", outputs=["{good_out}"])
        """,
    )
    rc = main(["run", "-k", "*"])
    assert rc == 1
    assert good_out.exists()


def test_validation_error_cycle(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    _write_recipe(
        project,
        """\
        from cook import get_context, Task
        ctx = get_context()
        a = Task(name="a", inputs=[], outputs=["a.txt"])
        b = Task(name="b", inputs=[a], outputs=["b.txt"])
        a.inputs.append(b)
        ctx.register(a)
        ctx.register(b)
        """,
    )
    rc = main(["run", "*"])
    assert rc == 1
    captured = capsys.readouterr().err
    assert "cycle" in captured.lower()


def test_recipe_import_error(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    _write_recipe(
        project,
        """\
        this is not valid python!!!
        """,
    )
    rc = main(["run", "*"])
    assert rc == 1
    captured = capsys.readouterr().err
    assert "error: loading recipe" in captured


def test_recipe_not_found(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    # No recipe.py created
    rc = main(["run", "*"])
    assert rc == 1
    captured = capsys.readouterr().err
    assert "error: loading recipe" in captured


def test_exit_code_success(project: Path) -> None:
    outfile = project / "out.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="ok", cmd="echo x > {outfile}", outputs=["{outfile}"])
        """,
    )
    assert main(["run", "ok"]) == 0


def test_exit_code_failure(project: Path) -> None:
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="fail", cmd="exit 1", outputs=["nope.txt"])
        """,
    )
    assert main(["run", "fail"]) == 1


def test_no_subcommand(capsys: pytest.CaptureFixture[str]) -> None:
    rc = main([])
    assert rc == 1


def test_jobs_flag(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    outfile = project / "out.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="t", cmd="echo ok > {outfile}", outputs=["{outfile}"])
        """,
    )
    rc = main(["run", "-j", "2", "t"])
    assert rc == 0
    assert outfile.exists()


def test_dry_run_does_not_create_store(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    """Dry-run should not create or modify the .cook.db file."""
    outfile = project / "out.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="build", cmd="echo x > {outfile}", outputs=["{outfile}"])
        """,
    )
    db_path = project / ".cook/store.db"
    assert not db_path.exists()
    rc = main(["run", "--dry-run", "build"])
    assert rc == 0
    assert not db_path.exists()


def test_dry_run_does_not_modify_store(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    """Dry-run with existing store should not modify any records."""
    outfile = project / "out.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="build", cmd="echo x > {outfile}", outputs=["{outfile}"])
        """,
    )
    # Build first to populate store
    rc = main(["run", "build"])
    assert rc == 0
    capsys.readouterr()

    # Record the store state
    from cook.store.sqlite import SqliteBuildStore

    with SqliteBuildStore(str(project / ".cook/store.db")) as store:
        record_before = store.get("build")
    assert record_before is not None

    # Modify input to make task stale, then dry-run
    outfile.unlink()
    rc = main(["run", "--dry-run", "build"])
    assert rc == 0
    assert "STALE" in capsys.readouterr().err

    # Store record should be unchanged
    with SqliteBuildStore(str(project / ".cook/store.db")) as store:
        record_after = store.get("build")
    assert record_after is not None
    assert record_after.digest == record_before.digest
    assert record_after.last_started == record_before.last_started


def test_dry_run_short_flag(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    outfile = project / "out.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="build", cmd="echo x > {outfile}", outputs=["{outfile}"])
        """,
    )
    rc = main(["run", "-n", "build"])
    assert rc == 0
    captured = capsys.readouterr().err
    assert "STALE (would run)" in captured
    assert not outfile.exists()


def test_unknown_executor_flag(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="t", cmd="true", outputs=[])
        """,
    )
    rc = main(["run", "--executor", "nosuch", "t"])
    assert rc == 1
    assert "unknown executor" in capsys.readouterr().err.lower()


def test_executor_short_flag(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="t", cmd="true", outputs=[])
        """,
    )
    rc = main(["run", "-x", "nosuch", "t"])
    assert rc == 1
    assert "unknown executor" in capsys.readouterr().err.lower()


def test_unknown_executor_config(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    (project / "cook.toml").write_text('[cook]\nexecutor = "nosuch"\n')
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="t", cmd="true", outputs=[])
        """,
    )
    rc = main(["run", "t"])
    assert rc == 1
    assert "unknown executor" in capsys.readouterr().err.lower()


def test_exec_slurm_executor_config(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    """CLI instantiates SlurmExecutor with config from cook.toml."""
    (project / "cook.toml").write_text(
        '[cook]\nexecutor = "slurm"\n\n'
        "[cook.slurm]\nmax_concurrent = 16\npoll_interval = 1.5\npoll_timeout = 7200\n"
    )
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="t", cmd="true", outputs=[])
        """,
    )
    # sbatch is not available locally, so task fails — but the slurm branch is exercised
    rc = main(["run", "t"])
    assert rc == 1
    output = capsys.readouterr().err.lower()
    assert "sbatch" in output


def test_exec_with_dependencies(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    """Pattern matches only leaf task, but dependencies run too."""
    out_dep = project / "dep.txt"
    out_leaf = project / "leaf.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        dep = ctx.sh(name="dep-task", cmd="echo dep > {out_dep}", outputs=["{out_dep}"])
        ctx.sh(name="leaf-task", cmd="echo leaf > {out_leaf}", inputs=[dep], outputs=["{out_leaf}"])
        """,
    )
    rc = main(["run", "leaf-task"])
    assert rc == 0
    assert out_dep.exists()
    assert out_leaf.exists()


def test_config_error(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    (project / "cook.toml").write_text("this is not valid toml {{{")
    _write_recipe(project, "from cook import get_context\nctx = get_context()\n")
    rc = main(["run", "*"])
    assert rc == 1
    captured = capsys.readouterr().err
    assert "error" in captured.lower()


def test_dry_run_with_deps(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    out_a = project / "a.txt"
    out_b = project / "b.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        a = ctx.sh(name="dep", cmd="echo a > {out_a}", outputs=["{out_a}"])
        ctx.sh(name="leaf", cmd="echo b > {out_b}", inputs=[a], outputs=["{out_b}"])
        """,
    )
    rc = main(["run", "--dry-run", "leaf"])
    assert rc == 0
    captured = capsys.readouterr().err
    assert "[dep] STALE (would run)" in captured
    assert "[leaf] STALE (would run)" in captured


def test_invalidate_no_matches(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="task1", cmd="true", outputs=[])
        """,
    )
    rc = main(["invalidate", "nonexistent-*"])
    assert rc == 1
    assert "matched no tasks" in capsys.readouterr().err


def test_inspect_no_pattern_no_default(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="task1", cmd="true", outputs=[])
        """,
    )
    rc = main(["inspect"])
    assert rc == 1
    assert "No target pattern" in capsys.readouterr().err


def test_inspect_recipe_error(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    _write_recipe(project, "def\n")
    rc = main(["inspect", "*"])
    assert rc == 1
    assert "error: loading recipe" in capsys.readouterr().err


def test_recipe_runtime_error(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    _write_recipe(project, "raise RuntimeError('boom')\n")
    rc = main(["run", "*"])
    assert rc == 1
    assert "error: loading recipe" in capsys.readouterr().err


def test_recipe_name_error(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    _write_recipe(project, "undefined_variable\n")
    rc = main(["run", "*"])
    assert rc == 1
    assert "error: loading recipe" in capsys.readouterr().err


def test_invalidate_recipe_error(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    _write_recipe(project, "def\n")
    rc = main(["invalidate", "*"])
    assert rc == 1
    assert "error: loading recipe" in capsys.readouterr().err


def test_exec_no_output_task(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    """Tasks with no outputs always run."""
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="always-run", cmd="true")
        """,
    )
    rc = main(["run", "always-run"])
    assert rc == 0
    captured = capsys.readouterr().err
    assert "Cooked" in captured


def test_dry_run_no_output_task(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    """Dry run shows no-output tasks as stale."""
    outfile = project / "out.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        dep = ctx.sh(name="always-run", cmd="true")
        ctx.sh(name="leaf", cmd="echo x > {outfile}", inputs=[dep], outputs=["{outfile}"])
        """,
    )
    # First run to populate store (creates .cook.db)
    rc = main(["run", "*"])
    assert rc == 0
    capsys.readouterr()

    # Dry run: always-run has no outputs -> stale, leaf depends on stale dep -> stale
    rc = main(["run", "--dry-run", "*"])
    assert rc == 0
    captured = capsys.readouterr().err
    assert "[always-run] STALE (would run)" in captured
    assert "[leaf] STALE (would run)" in captured


def test_dry_run_missing_output(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    """Dry run detects missing output files."""
    outfile = project / "out.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="build", cmd="echo x > {outfile}", outputs=["{outfile}"])
        """,
    )
    # Run to populate store
    rc = main(["run", "build"])
    assert rc == 0
    capsys.readouterr()

    # Delete the output
    outfile.unlink()

    # Dry run should detect it's stale
    rc = main(["run", "--dry-run", "build"])
    assert rc == 0
    captured = capsys.readouterr().err
    assert "STALE (would run)" in captured


def test_dry_run_staleness_cache(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    """Diamond deps: staleness cache is hit for shared dependency."""
    out_d = project / "d.txt"
    out_b = project / "b.txt"
    out_c = project / "c.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        d = ctx.sh(name="shared", cmd="echo d > {out_d}", outputs=["{out_d}"])
        ctx.sh(name="left", cmd="echo b > {out_b}", inputs=[d], outputs=["{out_b}"])
        ctx.sh(name="right", cmd="echo c > {out_c}", inputs=[d], outputs=["{out_c}"])
        """,
    )
    rc = main(["run", "--dry-run", "*"])
    assert rc == 0
    captured = capsys.readouterr().err
    assert "[shared] STALE (would run)" in captured


def test_load_recipe_empty_sys_path(
    project: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Cover the empty sys.path branch."""
    from cook.cli.util import load_recipe

    _write_recipe(
        project,
        """\
        # empty recipe
        """,
    )
    original_path = sys.path.copy()
    sys.path.clear()
    try:
        with Context():
            load_recipe(str(project / "recipe.py"))
    finally:
        sys.path.clear()
        sys.path.extend(original_path)


def test_main_module_block() -> None:
    """Cover the if __name__ == '__main__' block via runpy."""

    result = subprocess.run(
        [sys.executable, "-m", "cook.cli"],
        capture_output=True,
        text=True,
    )
    # Should print help and exit 1 (no subcommand)
    assert result.returncode == 1


def test_load_recipe_spec_none(
    project: Path,
) -> None:
    """Cover the spec is None branch."""
    from cook.cli.util import load_recipe

    # A directory can't be loaded as a module
    with pytest.raises(ImportError, match="Cannot load recipe"):
        load_recipe(str(project))


def test_dry_run_no_record_in_store(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    """Store exists but task has no record -> stale."""
    outfile = project / "out.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="new-task", cmd="echo x > {outfile}", outputs=["{outfile}"])
        """,
    )
    # Create a store with some other task so .cook.db exists
    from cook.store.sqlite import SqliteBuildStore

    store = SqliteBuildStore(str(project / ".cook/store.db"))
    store.close()

    rc = main(["run", "--dry-run", "new-task"])
    assert rc == 0
    captured = capsys.readouterr().err
    assert "STALE (would run)" in captured


def test_main_as_script(project: Path) -> None:
    """Cover the if __name__ == '__main__' block."""
    # We can't easily cover this in-process, so just verify the module attribute
    import cook.cli

    assert hasattr(cook.cli, "main")


def test_dry_run_detects_modified_source(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    """After modifying a source file, dry-run should show the task as stale."""
    infile = project / "src.txt"
    infile.write_text("v1")
    outfile = project / "out.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="build", cmd="cat {infile} > {outfile}", inputs=["{infile}"], outputs=["{outfile}"])
        """,
    )
    # Build successfully
    rc = main(["run", "build"])
    assert rc == 0
    assert outfile.read_text().strip() == "v1"
    capsys.readouterr()

    # Modify the source file
    infile.write_text("v2")

    # Dry run should detect staleness
    rc = main(["run", "--dry-run", "build"])
    assert rc == 0
    captured = capsys.readouterr().err
    assert "STALE (would run)" in captured


def test_validate_bad_recipe(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    """Validate with a broken recipe prints an error."""
    _write_recipe(project, "raise RuntimeError('boom')")
    rc = main(["validate", "*"])
    assert rc == 1
    assert "error: loading recipe" in capsys.readouterr().err  # validate bad recipe


def test_validate_marks_task_up_to_date(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    """Validate stores the effective digest so the task is up-to-date."""
    outfile = project / "out.txt"
    outfile.write_text("already built")
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="build", cmd="echo x > {outfile}", outputs=["{outfile}"])
        """,
    )
    rc = main(["validate", "build"])
    assert rc == 0
    captured = capsys.readouterr().err
    assert "[build] validated" in captured

    # Now dry-run should show up-to-date
    rc = main(["run", "--dry-run", "build"])
    assert rc == 0
    captured = capsys.readouterr().err
    assert "up-to-date" in captured


def test_validate_skips_no_output_task(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    """Tasks with no outputs cannot be validated."""
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="check", cmd="true")
        """,
    )
    rc = main(["validate", "check"])
    assert rc == 0
    captured = capsys.readouterr().err
    assert "[check] skipped" in captured
    assert "no outputs" in captured


def test_validate_skips_missing_outputs(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    """Validate skips tasks whose output files don't exist."""
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="build", cmd="echo x", outputs=["{project / "ghost.txt"}"])
        """,
    )
    rc = main(["validate", "build"])
    assert rc == 0
    captured = capsys.readouterr().err
    assert "[build] skipped" in captured
    assert "missing outputs" in captured


def test_validate_skips_always_run_dependency(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    """Tasks depending on always-run tasks cannot be validated."""
    outfile = project / "out.txt"
    outfile.write_text("done")
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        check = ctx.sh(name="check", cmd="true")
        ctx.sh(name="build", cmd="true", inputs=[check], outputs=["{outfile}"])
        """,
    )
    rc = main(["validate", "build"])
    assert rc == 0
    captured = capsys.readouterr().err
    assert "[build] skipped" in captured
    assert "always-run" in captured


def test_validate_with_deps(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    """Validate processes deps first so dependent digests are correct."""
    dep_out = project / "dep.txt"
    dep_out.write_text("dep done")
    main_out = project / "main.txt"
    main_out.write_text("main done")
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        dep = ctx.sh(name="dep", cmd="true", outputs=["{dep_out}"])
        ctx.sh(
            name="main",
            cmd="true",
            inputs=[dep],
            outputs=["{main_out}"],
        )
        """,
    )
    rc = main(["validate", "main"])
    assert rc == 0
    captured = capsys.readouterr().err
    assert "[dep] validated" in captured
    assert "[main] validated" in captured

    # Both should be up-to-date now
    rc = main(["run", "--dry-run", "*"])
    assert rc == 0
    captured = capsys.readouterr().err
    assert "up-to-date" in captured
    assert "STALE" not in captured


# --- ls command ---


def test_ls_lists_all_tasks(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="alpha", cmd="true")
        ctx.sh(name="beta", cmd="true")
        ctx.sh(name="gamma", cmd="true")
        """,
    )
    rc = main(["list"])
    assert rc == 0
    lines = capsys.readouterr().out.strip().splitlines()
    assert sorted(lines) == ["alpha", "beta", "gamma"]


def test_ls_with_pattern(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="build-a", cmd="true")
        ctx.sh(name="build-b", cmd="true")
        ctx.sh(name="test-a", cmd="true")
        """,
    )
    rc = main(["list", "build-*"])
    assert rc == 0
    lines = capsys.readouterr().out.strip().splitlines()
    assert sorted(lines) == ["build-a", "build-b"]


def test_ls_with_regex(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="build-a", cmd="true")
        ctx.sh(name="build-b", cmd="true")
        ctx.sh(name="test-a", cmd="true")
        """,
    )
    rc = main(["list", "--re", "^build-"])
    assert rc == 0
    lines = capsys.readouterr().out.strip().splitlines()
    assert sorted(lines) == ["build-a", "build-b"]


def test_ls_stale_filter(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    outfile = project / "out.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="has-out", cmd="echo x > {outfile}", outputs=["{outfile}"])
        ctx.sh(name="no-out", cmd="true")
        """,
    )
    # Run to make has-out current
    rc = main(["run", "has-out"])
    assert rc == 0
    capsys.readouterr()

    rc = main(["list", "--stale"])
    assert rc == 0
    lines = capsys.readouterr().out.strip().splitlines()
    # no-out has no outputs, always stale; has-out was just built
    assert "no-out" in lines
    assert "has-out" not in lines


def test_ls_current_filter(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    outfile = project / "out.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="has-out", cmd="echo x > {outfile}", outputs=["{outfile}"])
        ctx.sh(name="no-out", cmd="true")
        """,
    )
    # Run to make has-out current
    rc = main(["run", "has-out"])
    assert rc == 0
    capsys.readouterr()

    rc = main(["list", "--current"])
    assert rc == 0
    lines = capsys.readouterr().out.strip().splitlines()
    assert "has-out" in lines
    assert "no-out" not in lines


def test_ls_stale_no_store(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    """Without a store, --stale lists all tasks."""
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="a", cmd="true")
        ctx.sh(name="b", cmd="true")
        """,
    )
    rc = main(["list", "--stale"])
    assert rc == 0
    lines = capsys.readouterr().out.strip().splitlines()
    assert sorted(lines) == ["a", "b"]


def test_ls_current_no_store(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    """Without a store, --current lists nothing."""
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="a", cmd="true")
        """,
    )
    rc = main(["list", "--current"])
    assert rc == 0
    captured = capsys.readouterr().out.strip()
    assert captured == ""


def test_ls_recipe_error(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    _write_recipe(project, "raise RuntimeError('boom')")
    rc = main(["list"])
    assert rc == 1
    assert "error: loading recipe" in capsys.readouterr().err  # ls recipe error


def test_ls_pattern_no_match(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="task1", cmd="true")
        """,
    )
    rc = main(["list", "nonexistent-*"])
    assert rc == 1
    assert "matched no tasks" in capsys.readouterr().err


# --- -f / --file flag ---


def test_file_flag(project: Path) -> None:
    """The -f flag overrides the recipe file."""
    outfile = project / "out.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="custom", cmd="echo ok > {outfile}", outputs=["{outfile}"])
        """,
        name="custom.py",
    )
    rc = main(["-f", "custom.py", "run", "custom"])
    assert rc == 0
    assert outfile.exists()


def test_file_long_flag(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="t", cmd="true")
        """,
        name="other.py",
    )
    rc = main(["--file", "other.py", "list"])
    assert rc == 0
    assert "t" in capsys.readouterr().out


def test_file_flag_inspect(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="t", cmd="true")
        """,
        name="alt.py",
    )
    rc = main(["-f", "alt.py", "inspect", "t"])
    assert rc == 0
    assert "[t]" in capsys.readouterr().out


def test_file_flag_invalidate(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    outfile = project / "out.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="t", cmd="echo x > {outfile}", outputs=["{outfile}"])
        """,
        name="alt.py",
    )
    # Run first to have something to invalidate
    rc = main(["-f", "alt.py", "run", "t"])
    assert rc == 0
    capsys.readouterr()

    rc = main(["-f", "alt.py", "invalidate", "t"])
    assert rc == 0
    assert "Invalidated" in capsys.readouterr().err


def test_file_flag_validate(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    outfile = project / "out.txt"
    outfile.write_text("done")
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="t", cmd="true", outputs=["{outfile}"])
        """,
        name="alt.py",
    )
    rc = main(["-f", "alt.py", "validate", "t"])
    assert rc == 0
    assert "validated" in capsys.readouterr().err


def test_file_flag_missing(capsys: pytest.CaptureFixture[str]) -> None:
    rc = main(["-f", "nonexistent.py", "list"])
    assert rc == 1
    assert "error: loading recipe" in capsys.readouterr().err  # file flag missing


# --- -c / --config flag ---


def test_config_flag(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    """The -c flag loads config from a custom file."""
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="t", cmd="true")
        """,
        name="my_recipe.py",
    )
    config = project / "dev.toml"
    config.write_text('[cook]\nrecipe = "my_recipe.py"\n')
    rc = main(["-c", str(config), "list"])
    assert rc == 0
    assert "t" in capsys.readouterr().out


def test_config_flag_missing(capsys: pytest.CaptureFixture[str]) -> None:
    rc = main(["-c", "nonexistent.toml", "list"])
    assert rc == 1
    assert "Config file not found" in capsys.readouterr().err


def test_config_flag_file_overrides(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    """-f takes precedence over recipe in config file."""
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="from-flag", cmd="true")
        """,
        name="flag.py",
    )
    config = project / "dev.toml"
    config.write_text('[cook]\nrecipe = "wrong.py"\n')
    rc = main(["-c", str(config), "-f", "flag.py", "list"])
    assert rc == 0
    assert "from-flag" in capsys.readouterr().out


# --- verbosity and color flags ---


def test_verbose_flag(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    outfile = project / "out.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="t", cmd="echo ok > {outfile}", outputs=["{outfile}"])
        """,
    )
    rc = main(["-v", "run", "t"])
    assert rc == 0
    err = capsys.readouterr().err
    # Verbose mode shows the command
    assert f"$ echo ok > {outfile}" in err


def test_quiet_flag(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    outfile = project / "out.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="t", cmd="echo ok > {outfile}", outputs=["{outfile}"])
        """,
    )
    rc = main(["-q", "run", "t"])
    assert rc == 0
    err = capsys.readouterr().err
    # Quiet mode: no "Cooked" line, but summary still shows
    assert "Cooked" not in err


def test_verbose_and_quiet_conflict(capsys: pytest.CaptureFixture[str]) -> None:
    rc = main(["-v", "-q", "list"])
    assert rc == 1
    assert "Cannot use --verbose and --quiet" in capsys.readouterr().err


def test_color_never(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="t", cmd="true")
        """,
    )
    rc = main(["--color", "never", "list"])
    assert rc == 0


def test_color_always(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="t", cmd="true")
        """,
    )
    rc = main(["--color", "always", "list"])
    assert rc == 0


def test_stream_flag(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    outfile = project / "out.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="t", cmd="echo ok > {outfile}", outputs=["{outfile}"])
        """,
    )
    rc = main(["run", "-s", "t"])
    assert rc == 0


# --- build command ---


def test_build_by_output(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    outfile = project / "result.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="gen", cmd="echo ok > {outfile}", outputs=["{outfile}"])
        """,
    )
    rc = main(["build", str(outfile)])
    assert rc == 0
    assert outfile.exists()


def test_build_glob_pattern(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    out1 = project / "a.o"
    out2 = project / "b.o"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="compile-a", cmd="echo a > {out1}", outputs=["{out1}"])
        ctx.sh(name="compile-b", cmd="echo b > {out2}", outputs=["{out2}"])
        ctx.sh(name="unrelated", cmd="true")
        """,
    )
    rc = main(["build", "*.o"])
    assert rc == 0
    assert out1.exists()
    assert out2.exists()
    err = capsys.readouterr().err
    # Both compile tasks ran, unrelated did not
    assert "compile-a" in err
    assert "compile-b" in err


def test_build_regex(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    outfile = project / "output.dat"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="gen", cmd="echo ok > {outfile}", outputs=["{outfile}"])
        """,
    )
    rc = main(["build", "-r", r"\.dat$"])
    assert rc == 0
    assert outfile.exists()


def test_build_no_match(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="t", cmd="true", outputs=["foo.txt"])
        """,
    )
    rc = main(["build", "*.xyz"])
    assert rc == 1
    assert "matched no task outputs" in capsys.readouterr().err


def test_build_no_pattern(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="t", cmd="true")
        """,
    )
    rc = main(["build"])
    assert rc == 1
    assert "No output pattern" in capsys.readouterr().err


def test_build_dry_run(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    outfile = project / "out.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="gen", cmd="echo ok > {outfile}", outputs=["{outfile}"])
        """,
    )
    rc = main(["build", "-n", str(outfile)])
    assert rc == 0
    assert not outfile.exists()
    assert "STALE" in capsys.readouterr().err


def test_build_dry_run_with_store(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    outfile = project / "out.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="gen", cmd="echo ok > {outfile}", outputs=["{outfile}"])
        """,
    )
    # Build first to create the store
    rc = main(["build", str(outfile)])
    assert rc == 0
    capsys.readouterr()

    # Dry-run should show up-to-date
    rc = main(["build", "-n", str(outfile)])
    assert rc == 0
    assert "up-to-date" in capsys.readouterr().err


def test_build_regex_invalid(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="t", cmd="true", outputs=["foo.txt"])
        """,
    )
    rc = main(["build", "-r", "[invalid"])
    assert rc == 1
    assert "Invalid regex" in capsys.readouterr().err


def test_build_regex_no_match(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="t", cmd="true", outputs=["foo.txt"])
        """,
    )
    rc = main(["build", "-r", r"\.xyz$"])
    assert rc == 1
    assert "matched no task outputs" in capsys.readouterr().err


def test_build_multiple_patterns(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    out1 = project / "a.o"
    out2 = project / "b.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="t1", cmd="echo a > {out1}", outputs=["{out1}"])
        ctx.sh(name="t2", cmd="echo b > {out2}", outputs=["{out2}"])
        """,
    )
    rc = main(["build", "*.o", "*.txt"])
    assert rc == 0
    assert out1.exists()
    assert out2.exists()


def test_build_one_pattern_no_match(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="t", cmd="true", outputs=["foo.txt"])
        """,
    )
    rc = main(["build", "*.txt", "*.xyz"])
    assert rc == 1
    assert "*.xyz" in capsys.readouterr().err


# --- --json flag ---


def test_ls_json(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="alpha", cmd="true")
        ctx.sh(name="beta", cmd="true")
        """,
    )
    rc = main(["list", "--json"])
    assert rc == 0
    import json

    lines = capsys.readouterr().out.strip().splitlines()
    assert len(lines) == 2
    objs = [json.loads(line) for line in lines]
    names = {o["name"] for o in objs}
    assert names == {"alpha", "beta"}


def test_ls_json_with_stale(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="t", cmd="true")
        """,
    )
    rc = main(["list", "--json", "--stale"])
    assert rc == 0
    import json

    lines = capsys.readouterr().out.strip().splitlines()
    obj = json.loads(lines[0])
    assert obj["name"] == "t"
    assert obj["stale"] is True


def test_inspect_json(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    outfile = project / "out.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="gen", cmd="echo ok", outputs=["{outfile}"])
        """,
    )
    rc = main(["inspect", "--json", "gen"])
    assert rc == 0
    import json

    lines = capsys.readouterr().out.strip().splitlines()
    obj = json.loads(lines[0])
    assert obj["name"] == "gen"
    assert obj["stale"] is True
    assert obj["cmd"] == "echo ok"
    assert str(outfile) in obj["outputs"]


def test_inspect_json_with_history(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    outfile = project / "out.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="gen", cmd="echo ok > {outfile}", outputs=["{outfile}"])
        """,
    )
    # Run first to create history
    rc = main(["run", "gen"])
    assert rc == 0
    capsys.readouterr()

    rc = main(["inspect", "--json", "gen"])
    assert rc == 0
    import json

    lines = capsys.readouterr().out.strip().splitlines()
    obj = json.loads(lines[0])
    assert obj["name"] == "gen"
    assert obj["stale"] is False
    assert "history" in obj
    assert "last_succeeded" in obj["history"]
    assert "duration" in obj["history"]


def test_inspect_json_failed_task(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    outfile = project / "out.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        ctx.sh(name="fail", cmd="exit 1", outputs=["{outfile}"])
        """,
    )
    # Run to create a failure record
    rc = main(["run", "fail"])
    assert rc == 1
    capsys.readouterr()

    rc = main(["inspect", "--json", "fail"])
    assert rc == 0
    import json

    lines = capsys.readouterr().out.strip().splitlines()
    obj = json.loads(lines[0])
    assert obj["name"] == "fail"
    assert "history" in obj
    assert "last_failed" in obj["history"]
    assert "error" in obj["history"]


# --- group command ---


def test_group_run(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    out_a = project / "a.txt"
    out_b = project / "b.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        with ctx.group("all") as g:
            ctx.sh(name="a", cmd="echo a > {out_a}", outputs=["{out_a}"])
            ctx.sh(name="b", cmd="echo b > {out_b}", outputs=["{out_b}"])
        """,
    )
    rc = main(["run", "all"])
    assert rc == 0
    assert out_a.exists()
    assert out_b.exists()
    err = capsys.readouterr().err
    assert "a" in err
    assert "b" in err


def test_group_incremental(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    out_a = project / "a.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        with ctx.group("all") as g:
            ctx.sh(name="a", cmd="echo a > {out_a}", outputs=["{out_a}"])
        """,
    )
    # First run
    rc = main(["run", "all"])
    assert rc == 0
    capsys.readouterr()

    # Second run — should be fresh
    rc = main(["run", "all"])
    assert rc == 0
    err = capsys.readouterr().err
    assert "Fresh" in err


def test_group_inspect(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    """Inspect should show group as up-to-date after a successful run."""
    out_a = project / "a.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        with ctx.group("grp") as g:
            ctx.sh(name="a", cmd="echo a > {out_a}", outputs=["{out_a}"])
        """,
    )
    rc = main(["run", "grp"])
    assert rc == 0
    capsys.readouterr()

    rc = main(["inspect", "grp"])
    assert rc == 0
    out = capsys.readouterr().out
    assert "up-to-date" in out
    # Should NOT say "missing" for the virtual output
    assert "missing" not in out


def test_group_inspect_stale(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    """Inspect should show group as stale when never run."""
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        with ctx.group("grp") as g:
            ctx.sh(name="a", cmd="true", outputs=["a.txt"])
        """,
    )
    rc = main(["inspect", "grp"])
    assert rc == 0
    out = capsys.readouterr().out
    assert "STALE" in out
    assert "never run" in out


def test_group_inspect_json(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    """JSON inspect should include group task."""
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        with ctx.group("grp") as g:
            ctx.sh(name="a", cmd="true", outputs=["a.txt"])
        """,
    )
    rc = main(["inspect", "--json", "grp"])
    assert rc == 0
    import json

    lines = capsys.readouterr().out.strip().splitlines()
    objs = {json.loads(line)["name"]: json.loads(line) for line in lines}
    assert "grp" in objs
    assert objs["grp"]["type"] == "GroupTask"


def test_group_validate(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    """Validate should be able to mark a group as up-to-date after running."""
    out_a = project / "a.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        with ctx.group("grp") as g:
            ctx.sh(name="a", cmd="echo a > {out_a}", outputs=["{out_a}"])
        """,
    )
    # Run first to create marker and outputs
    rc = main(["run", "grp"])
    assert rc == 0
    capsys.readouterr()

    # Invalidate, then validate
    rc = main(["invalidate", "grp"])
    assert rc == 0
    capsys.readouterr()

    rc = main(["validate", "grp"])
    assert rc == 0
    err = capsys.readouterr().err
    assert "grp] validated" in err


def test_group_list_stale(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    """List --stale should include group when children are stale."""
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        with ctx.group("grp") as g:
            ctx.sh(name="a", cmd="true", outputs=["a.txt"])
        """,
    )
    rc = main(["list", "--stale"])
    assert rc == 0
    out = capsys.readouterr().out
    assert "grp" in out


def test_group_list_current(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    """List --current should include group when all children are fresh."""
    out_a = project / "a.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        with ctx.group("grp") as g:
            ctx.sh(name="a", cmd="echo a > {out_a}", outputs=["{out_a}"])
        """,
    )
    rc = main(["run", "grp"])
    assert rc == 0
    capsys.readouterr()

    rc = main(["list", "--current"])
    assert rc == 0
    out = capsys.readouterr().out
    assert "grp" in out


def test_group_dry_run(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    """Dry run should show group and children."""
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        with ctx.group("grp") as g:
            ctx.sh(name="a", cmd="true", outputs=["a.txt"])
        """,
    )
    rc = main(["run", "-n", "grp"])
    assert rc == 0
    err = capsys.readouterr().err
    assert "grp" in err
    assert "STALE" in err


def test_group_invalidate(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    """Invalidate on a group should force re-run."""
    out_a = project / "a.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        with ctx.group("grp") as g:
            ctx.sh(name="a", cmd="echo a > {out_a}", outputs=["{out_a}"])
        """,
    )
    rc = main(["run", "grp"])
    assert rc == 0
    capsys.readouterr()

    rc = main(["invalidate", "grp"])
    assert rc == 0
    capsys.readouterr()

    # Should re-run after invalidation
    rc = main(["run", "grp"])
    assert rc == 0
    err = capsys.readouterr().err
    assert "Cooked" in err


def test_group_empty(project: Path, capsys: pytest.CaptureFixture[str]) -> None:
    """Empty group should run successfully."""
    _write_recipe(
        project,
        """\
        from cook import get_context
        ctx = get_context()
        with ctx.group("empty") as g:
            pass
        """,
    )
    rc = main(["run", "empty"])
    assert rc == 0


def test_group_child_change_propagates(
    project: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    """When a child's input changes, the group and its dependents should re-run."""
    infile = project / "src.txt"
    infile.write_text("original")
    out_a = project / "a.txt"
    out_b = project / "b.txt"
    _write_recipe(
        project,
        f"""\
        from cook import get_context
        ctx = get_context()
        with ctx.group("grp") as g:
            ctx.sh(name="a", cmd="cat {infile} > {out_a}", inputs=["{infile}"], outputs=["{out_a}"])
        ctx.sh(name="downstream", cmd="cat {out_a} > {out_b}", inputs=[g], outputs=["{out_b}"])
        """,
    )
    rc = main(["run", "downstream"])
    assert rc == 0
    capsys.readouterr()

    # Modify input
    infile.write_text("changed")
    rc = main(["run", "downstream"])
    assert rc == 0
    err = capsys.readouterr().err
    # Both should re-run
    assert "Cooked" in err



from __future__ import annotations

from pathlib import Path

import pytest

from cook.task import ShellTask, Task
from cook.transform import (
    check_cycles,
    check_deps_registered,
    check_extra_keys,
    check_outputs,
    resolve_file_deps,
)


def _tasks(*task_list: Task) -> dict[str, Task]:
    return {t.name: t for t in task_list}


# --- check_deps_registered ---


def test_check_deps_registered_valid() -> None:
    a = Task(name="a")
    b = Task(name="b", inputs=[a])
    check_deps_registered(_tasks(a, b))


def test_check_deps_registered_missing() -> None:
    ghost = Task(name="ghost")
    a = Task(name="a", inputs=[ghost])
    with pytest.raises(ValueError, match="not registered"):
        check_deps_registered(_tasks(a))


# --- check_outputs ---


def test_check_outputs_valid() -> None:
    a = Task(name="a", outputs=["a.o"])
    b = Task(name="b", outputs=["b.o"])
    check_outputs(_tasks(a, b))


def test_check_outputs_duplicate() -> None:
    a = Task(name="a", outputs=["out.o"])
    b = Task(name="b", outputs=["out.o"])
    with pytest.raises(ValueError, match="Duplicate output path"):
        check_outputs(_tasks(a, b))


def test_check_outputs_duplicate_resolved(tmp_path: Path) -> None:
    p = str(tmp_path / "out.o")
    a = Task(name="a", outputs=[p])
    b = Task(name="b", outputs=[p])
    with pytest.raises(ValueError, match="Duplicate output path"):
        check_outputs(_tasks(a, b))


def test_check_outputs_input_overlaps_output() -> None:
    a = Task(name="a", inputs=["file.txt"], outputs=["file.txt"])
    with pytest.raises(ValueError, match="both an input and an output"):
        check_outputs(_tasks(a))


# --- resolve_file_deps ---


def test_resolve_adds_producer() -> None:
    a = ShellTask(name="compile", cmd="cc -c", outputs=["foo.o"])
    b = ShellTask(name="link", cmd="cc", inputs=["foo.o"], outputs=["app"])
    resolve_file_deps(_tasks(a, b))
    assert a in b.task_deps


def test_resolve_no_match() -> None:
    a = ShellTask(name="compile", cmd="cc -c", outputs=["foo.o"])
    b = ShellTask(name="link", cmd="cc", inputs=["bar.o"], outputs=["app"])
    resolve_file_deps(_tasks(a, b))
    assert a not in b.task_deps


def test_resolve_already_explicit_dep() -> None:
    a = ShellTask(name="compile", cmd="cc -c", outputs=["foo.o"])
    b = ShellTask(name="link", cmd="cc", inputs=[a, "foo.o"], outputs=["app"])
    resolve_file_deps(_tasks(a, b))
    # Should not add a duplicate
    assert b.task_deps == [a]


def test_resolve_preserves_file_input() -> None:
    a = ShellTask(name="compile", cmd="cc -c", outputs=["foo.o"])
    b = ShellTask(name="link", cmd="cc", inputs=["foo.o"], outputs=["app"])
    resolve_file_deps(_tasks(a, b))
    assert "foo.o" in b.file_inputs
    assert a in b.task_deps


def test_resolve_chain() -> None:
    a = ShellTask(name="a", cmd="a", outputs=["x.o"])
    b = ShellTask(name="b", cmd="b", inputs=["x.o"], outputs=["y.o"])
    c = ShellTask(name="c", cmd="c", inputs=["y.o"], outputs=["z.o"])
    resolve_file_deps(_tasks(a, b, c))
    assert a in b.task_deps
    assert b in c.task_deps


def test_resolve_diamond() -> None:
    a = ShellTask(name="a", cmd="a", outputs=["shared.o"])
    b = ShellTask(name="b", cmd="b", inputs=["shared.o"], outputs=["b.o"])
    c = ShellTask(name="c", cmd="c", inputs=["shared.o"], outputs=["c.o"])
    resolve_file_deps(_tasks(a, b, c))
    assert a in b.task_deps
    assert a in c.task_deps


# --- check_cycles ---


def test_check_cycles_valid() -> None:
    a = Task(name="a")
    b = Task(name="b", inputs=[a])
    check_cycles(_tasks(a, b))


def test_check_cycles_direct() -> None:
    a = Task(name="a")
    b = Task(name="b")
    a.inputs = [b]
    b.inputs = [a]
    with pytest.raises(ValueError, match="cycle"):
        check_cycles(_tasks(a, b))


def test_check_cycles_self() -> None:
    a = Task(name="a")
    a.inputs = [a]
    with pytest.raises(ValueError, match="cycle"):
        check_cycles(_tasks(a))


def test_check_cycles_diamond_no_false_positive() -> None:
    d = Task(name="d")
    b = Task(name="b", inputs=[d])
    c = Task(name="c", inputs=[d])
    a = Task(name="a", inputs=[b, c])
    check_cycles(_tasks(a, b, c, d))


# --- integration: resolve + cycle detection ---


def test_resolve_preserves_existing_deps() -> None:
    """Pre-populated deps are not overwritten by resolve_file_deps."""
    a = ShellTask(name="a", cmd="a")
    b = ShellTask(name="b", cmd="b", outputs=["x.o"])
    c = ShellTask(name="c", cmd="c", inputs=["x.o"])
    c._deps.add(a)
    resolve_file_deps(_tasks(a, b, c))
    assert a in c.task_deps
    assert b in c.task_deps


def test_resolve_idempotent() -> None:
    """Running resolve_file_deps twice gives the same result."""
    a = ShellTask(name="a", cmd="a", outputs=["x.o"])
    b = ShellTask(name="b", cmd="b", inputs=["x.o"])
    tasks = _tasks(a, b)
    resolve_file_deps(tasks)
    assert a in b.task_deps
    resolve_file_deps(tasks)
    assert a in b.task_deps
    assert b.task_deps == [a]


def test_resolve_creates_cycle() -> None:
    """A outputs x.o, B inputs x.o and A depends on B → cycle after resolution."""
    a = ShellTask(name="a", cmd="a", inputs=[], outputs=["x.o"])
    b = ShellTask(name="b", cmd="b", inputs=["x.o"])
    a.inputs = [b]  # explicit dep: a -> b
    # After resolution: b -> a (via x.o), creating a -> b -> a cycle
    resolve_file_deps(_tasks(a, b))
    with pytest.raises(ValueError, match="cycle"):
        check_cycles(_tasks(a, b))


# --- check_extra_keys ---


def test_extra_keys_valid_executor_name() -> None:
    a = ShellTask(name="a", cmd="echo", extra={"slurm": {"mem": "8G"}})
    check_extra_keys(_tasks(a))  # should not raise


def test_extra_keys_unknown_rejected() -> None:
    a = ShellTask(name="a", cmd="echo", extra={"bogus": {"x": "y"}})
    with pytest.raises(ValueError, match="unknown extra key.*bogus"):
        check_extra_keys(_tasks(a))


def test_extra_keys_empty_ok() -> None:
    a = ShellTask(name="a", cmd="echo")
    check_extra_keys(_tasks(a))  # should not raise



from __future__ import annotations

import argparse

from ..config import Config
from ..context import Context
from ..scheduler import compute_effective_digest
from ..store import TaskRecord
from ..store.sqlite import SqliteBuildStore
from ..transform import collect_transitive
from ..ui import Output
from .util import match_targets


def cmd_validate(
    args: argparse.Namespace, config: Config, ctx: Context, ui: Output
) -> int:
    targets = match_targets(ctx.tasks, args.pattern, config.default, args.regex)

    all_tasks = collect_transitive(targets)

    # Check if any tasks can actually be validated before creating the db
    validatable = []
    for task in all_tasks:
        if not task.outputs:
            ui.status(f"[{task.name}] skipped (no outputs or always-run dependency)")
            continue
        missing = [
            ctx.resolve_to_resource(o).label
            for o in task.outputs
            if not ctx.resolve_to_resource(o).exists()
        ]
        if missing:
            paths = ", ".join(missing)
            ui.status(f"[{task.name}] skipped (missing outputs: {paths})")
            continue
        validatable.append(task)

    if not validatable:
        return 0

    with SqliteBuildStore(str(ctx.db_path)) as store:
        for task in validatable:
            effective = compute_effective_digest(
                task, store, project_root=ctx.project_root
            )
            if effective is None:
                ui.status(f"[{task.name}] skipped (always-run dependency)")
                continue
            store.save(TaskRecord(task_id=task.task_id, digest=effective))
            ui.status(f"[{task.name}] validated")

    return 0



from __future__ import annotations

import os
import sys
from enum import IntEnum


class Verbosity(IntEnum):
    QUIET = 0
    NORMAL = 1
    VERBOSE = 2


def _should_color(override: bool | None) -> bool:
    if override is not None:
        return override
    if os.environ.get("NO_COLOR"):
        return False
    if os.environ.get("CLICOLOR_FORCE"):
        return True
    return hasattr(sys.stderr, "isatty") and sys.stderr.isatty()


_RESET = "\033[0m"
_BOLD = "\033[1m"
_DIM = "\033[2m"
_YELLOW = "\033[33m"
_CYAN = "\033[36m"
_BOLD_RED = "\033[1;31m"
_BOLD_GREEN = "\033[1;32m"


class Style:
    def __init__(self, enabled: bool) -> None:
        self._enabled = enabled

    def _wrap(self, code: str, text: str) -> str:
        if not self._enabled:
            return text
        return f"{code}{text}{_RESET}"

    def green(self, text: str) -> str:
        return self._wrap(_BOLD_GREEN, text)

    def red(self, text: str) -> str:
        return self._wrap(_BOLD_RED, text)

    def yellow(self, text: str) -> str:
        return self._wrap(_YELLOW, text)

    def cyan(self, text: str) -> str:
        return self._wrap(_CYAN, text)

    def dim(self, text: str) -> str:
        return self._wrap(_DIM, text)

    def bold(self, text: str) -> str:
        return self._wrap(_BOLD, text)


class Output:
    def __init__(
        self,
        verbosity: Verbosity = Verbosity.NORMAL,
        color: bool | None = None,
    ) -> None:
        self.verbosity = verbosity
        self.style = Style(_should_color(color))
        self._total = 0
        self._done = 0

    def set_total(self, total: int) -> None:
        self._total = total

    def _counter(self) -> str:
        self._done += 1
        width = len(str(self._total))
        return f"[{self._done:>{width}}/{self._total}]"

    def task_fresh(self, name: str) -> None:
        if self.verbosity < Verbosity.NORMAL:
            return
        counter = self._counter()
        action = self.style.cyan("Fresh")
        print(f"{counter} {action}   {name}", file=sys.stderr)

    def task_cooked(self, name: str, elapsed: float) -> None:
        if self.verbosity < Verbosity.NORMAL:
            return
        counter = self._counter()
        action = self.style.green("Cooked")
        timing = self.style.dim(f"({elapsed:.1f}s)")
        print(f"{counter} {action}  {name} {timing}", file=sys.stderr)

    def task_failed(self, name: str, elapsed: float, message: str) -> None:
        # Always show failures, even in quiet mode
        counter = self._counter()
        action = self.style.red("FAILED")
        timing = self.style.dim(f"({elapsed:.1f}s)")
        print(f"{counter} {action}  {name} {timing}", file=sys.stderr)
        for line in message.splitlines():
            print(f"         {line}", file=sys.stderr)

    def task_skipped(self, name: str, failed_dep: str) -> None:
        if self.verbosity < Verbosity.NORMAL:
            return
        counter = self._counter()
        action = self.style.yellow("Skipped")
        detail = self.style.dim(f"(dep {failed_dep!r} failed)")
        print(f"{counter} {action} {name} {detail}", file=sys.stderr)

    def summary(
        self,
        cooked: int,
        fresh: int,
        failed: int,
        skipped: int,
        elapsed: float,
    ) -> None:
        parts: list[str] = []
        if cooked:
            parts.append(self.style.green(f"{cooked} cooked"))
        if fresh:
            parts.append(self.style.cyan(f"{fresh} fresh"))
        if skipped:
            parts.append(self.style.yellow(f"{skipped} skipped"))
        if failed:
            parts.append(self.style.red(f"{failed} failed"))

        summary = ", ".join(parts)
        timing = self.style.dim(f"in {elapsed:.1f}s")

        if failed:
            status = self.style.red("Build failed")
        else:
            status = self.style.green("Build finished")

        print(f"\n{status}: {summary} {timing}", file=sys.stderr)

    def error(self, msg: str) -> None:
        label = self.style.red("error")
        print(f"{label}: {msg}", file=sys.stderr)

    def status(self, msg: str) -> None:
        if self.verbosity >= Verbosity.NORMAL:
            print(msg, file=sys.stderr)

    def verbose(self, msg: str) -> None:
        if self.verbosity >= Verbosity.VERBOSE:
            print(msg, file=sys.stderr)



from __future__ import annotations

import asyncio
from dataclasses import dataclass
from pathlib import Path

import pytest

from cook.config import ConfigError
from cook.executor import Executor, LocalExecutor, TaskExecutionError
from cook.executor.local import LocalConfig
from cook.executor.slurm import SlurmConfig
from cook.task import ShellTask, Task


async def test_local_executor_runs_echo() -> None:
    executor = LocalExecutor()
    task = ShellTask(name="echo-test", cmd="echo hello")
    await executor.execute(task)


async def test_local_executor_captures_stdout(tmp_path: object) -> None:
    outfile = str(tmp_path) + "/out.txt"  # type: ignore[operator]
    executor = LocalExecutor()
    task = ShellTask(name="capture", cmd=f"echo hello > {outfile}")
    await executor.execute(task)
    assert Path(outfile).read_text().strip() == "hello"


async def test_local_executor_raises_on_nonzero_exit() -> None:
    executor = LocalExecutor()
    task = ShellTask(name="fail", cmd="exit 1")
    with pytest.raises(TaskExecutionError):
        await executor.execute(task)


async def test_task_execution_error_contains_details() -> None:
    executor = LocalExecutor()
    task = ShellTask(name="fail-detail", cmd="echo badstuff >&2; exit 42")
    with pytest.raises(TaskExecutionError) as exc_info:
        await executor.execute(task)
    err = exc_info.value
    assert err.task is task
    assert err.returncode == 42
    assert "badstuff" in err.stderr
    assert "fail-detail" in str(err)


async def test_semaphore_limits_concurrency() -> None:
    max_concurrent = 2
    executor = LocalExecutor(max_concurrent=max_concurrent)
    running = 0
    peak = 0

    original_handler = LocalExecutor._handlers[ShellTask]

    async def tracking_handler(ex: Executor, task: Task) -> None:
        nonlocal running, peak
        running += 1
        peak = max(peak, running)
        await asyncio.sleep(0.05)
        running -= 1

    LocalExecutor.register_handler(tracking_handler, task_type=ShellTask)
    try:
        tasks = [ShellTask(name=f"t{i}", cmd="true") for i in range(max_concurrent + 1)]
        await asyncio.gather(*(executor.execute(t) for t in tasks))
        assert peak <= max_concurrent
    finally:
        LocalExecutor.register_handler(original_handler, task_type=ShellTask)


async def test_mro_handler_resolution() -> None:
    @dataclass
    class MyShellTask(ShellTask):
        pass

    executor = LocalExecutor()
    task = MyShellTask(name="sub", cmd="echo mro")
    # Should use ShellTask handler via MRO
    await executor.execute(task)


async def test_unregistered_task_type_raises() -> None:
    @dataclass
    class CustomTask(Task):
        pass

    executor = LocalExecutor()
    task = CustomTask(name="custom")
    with pytest.raises(TypeError, match="No handler registered"):
        await executor.execute(task)


async def test_custom_handler() -> None:
    @dataclass
    class CustomTask(Task):
        value: str = ""

    called_with: list[str] = []

    async def handle_custom(executor: Executor, task: Task) -> None:
        assert isinstance(task, CustomTask)
        called_with.append(task.value)

    executor = LocalExecutor()
    LocalExecutor.register_handler(handle_custom, task_type=CustomTask)
    try:
        task = CustomTask(name="custom", value="hello")
        await executor.execute(task)
        assert called_with == ["hello"]
    finally:
        del LocalExecutor._handlers[CustomTask]


async def test_env_none_inherits_parent() -> None:
    executor = LocalExecutor()
    task = ShellTask(name="env-inherit", cmd="echo $HOME", env=None)
    # Should not raise -- inherits env and can run
    await executor.execute(task)


async def test_env_dict_replaces_environment(tmp_path: object) -> None:
    outfile = str(tmp_path) + "/env_out.txt"  # type: ignore[operator]
    executor = LocalExecutor()
    # Use a minimal env with PATH so shell can find echo
    task = ShellTask(
        name="env-replace",
        cmd=f"echo HOME=$HOME > {outfile}",
        env={"FOO": "bar", "PATH": "/usr/bin:/bin"},
    )
    await executor.execute(task)
    content = Path(outfile).read_text().strip()
    assert content == "HOME="


async def test_cwd_sets_working_directory(tmp_path: object) -> None:
    outfile = str(tmp_path) + "/cwd_out.txt"  # type: ignore[operator]
    executor = LocalExecutor()
    task = ShellTask(name="cwd-test", cmd=f"pwd > {outfile}", cwd=str(tmp_path))
    await executor.execute(task)
    result = Path(outfile).read_text().strip()
    # Resolve symlinks for macOS /private/var/... vs /var/...
    assert Path(result).resolve() == Path(str(tmp_path)).resolve()


async def test_cancel_kills_running_subprocess() -> None:
    """Cancelling a running task should terminate the subprocess promptly."""
    executor = LocalExecutor()
    task = ShellTask(name="long-running", cmd="sleep 60")
    coro = executor.execute(task)
    t = asyncio.ensure_future(coro)
    await asyncio.sleep(0.1)  # let subprocess start
    t.cancel()
    with pytest.raises(asyncio.CancelledError):
        await t
    # Give OS a moment to reap the process
    await asyncio.sleep(0.1)


async def test_stderr_invalid_utf8() -> None:
    """Binary stderr should not crash error reporting."""
    executor = LocalExecutor()
    # printf \xff produces an invalid UTF-8 byte
    task = ShellTask(name="bad-utf8", cmd="printf '\\xff' >&2; exit 1")
    with pytest.raises(TaskExecutionError) as exc_info:
        await executor.execute(task)
    assert exc_info.value.returncode == 1
    # Should contain the replacement character, not crash
    assert isinstance(exc_info.value.stderr, str)


async def test_stream_mode_passes_through() -> None:
    executor = LocalExecutor(stream=True)
    task = ShellTask(name="stream-ok", cmd="echo hello")
    await executor.execute(task)


async def test_stream_mode_failure() -> None:
    executor = LocalExecutor(stream=True)
    task = ShellTask(name="stream-fail", cmd="exit 1")
    with pytest.raises(TaskExecutionError) as exc_info:
        await executor.execute(task)
    assert exc_info.value.returncode == 1
    assert "streamed to terminal" in exc_info.value.stderr


async def test_local_executor_runs_cmd_sequence(tmp_path: Path) -> None:
    outfile = tmp_path / "out.txt"
    executor = LocalExecutor()
    # Sequence form: no shell expansion, args passed directly to exec
    task = ShellTask(
        name="seq", cmd=["python3", "-c", f"open('{outfile}', 'w').write('ok')"]
    )
    await executor.execute(task)
    assert outfile.read_text() == "ok"


async def test_local_executor_cmd_sequence_no_shell_expansion(tmp_path: Path) -> None:
    outfile = tmp_path / "out.txt"
    executor = LocalExecutor()
    # "$HOME" should NOT be expanded when using sequence form
    task = ShellTask(
        name="no-expand",
        cmd=["python3", "-c", f"open('{outfile}', 'w').write('$HOME')"],
    )
    await executor.execute(task)
    assert outfile.read_text() == "$HOME"


async def test_local_executor_cmd_sequence_raises_on_failure() -> None:
    executor = LocalExecutor()
    task = ShellTask(name="seq-fail", cmd=["python3", "-c", "raise SystemExit(2)"])
    with pytest.raises(TaskExecutionError) as exc_info:
        await executor.execute(task)
    assert exc_info.value.returncode == 2


async def test_local_executor_cmd_sequence_stream_mode(tmp_path: Path) -> None:
    outfile = tmp_path / "out.txt"
    executor = LocalExecutor(stream=True)
    task = ShellTask(
        name="seq-stream",
        cmd=["python3", "-c", f"open('{outfile}', 'w').write('streamed')"],
    )
    await executor.execute(task)
    assert outfile.read_text() == "streamed"


# --- LocalConfig ---


def test_local_config_defaults():
    cfg = LocalConfig()
    assert cfg.max_concurrent == 1


def test_local_config_valid():
    cfg = LocalConfig(max_concurrent=8)
    assert cfg.max_concurrent == 8


def test_local_config_invalid_type():
    with pytest.raises(ConfigError, match="integer"):
        LocalConfig(max_concurrent=True)


def test_local_config_too_low():
    with pytest.raises(ConfigError, match="must be >= 1"):
        LocalConfig(max_concurrent=0)


def test_local_from_config():
    executor = LocalExecutor.from_config({"max_concurrent": 4})
    assert executor._semaphore._value == 4


def test_local_from_config_empty():
    executor = LocalExecutor.from_config({})
    assert executor._semaphore._value == 1


def test_local_from_config_jobs_override():
    executor = LocalExecutor.from_config({"max_concurrent": 4}, jobs=8)
    assert executor._semaphore._value == 8


# --- SlurmConfig ---


def test_slurm_config_defaults():
    cfg = SlurmConfig()
    assert cfg.max_concurrent == 64
    assert cfg.poll_interval == 2.0
    assert cfg.poll_timeout == 86400.0
    assert cfg.poll_retries == 10
    assert cfg.defaults == {}


def test_slurm_config_valid():
    cfg = SlurmConfig(max_concurrent=32, poll_interval=5.0, defaults={"mem": "4G"})
    assert cfg.max_concurrent == 32
    assert cfg.poll_interval == 5.0
    assert cfg.defaults == {"mem": "4G"}


def test_slurm_config_max_concurrent_invalid_type():
    with pytest.raises(ConfigError, match="integer"):
        SlurmConfig(max_concurrent=True)


def test_slurm_config_max_concurrent_too_low():
    with pytest.raises(ConfigError, match="must be >= 1"):
        SlurmConfig(max_concurrent=0)


def test_slurm_config_poll_interval_invalid_type():
    with pytest.raises(ConfigError, match="number"):
        SlurmConfig(poll_interval=True)  # type: ignore[arg-type]


def test_slurm_config_poll_interval_too_low():
    with pytest.raises(ConfigError, match="must be > 0"):
        SlurmConfig(poll_interval=0)


def test_slurm_config_poll_interval_int_coerced():
    cfg = SlurmConfig(poll_interval=3)  # type: ignore[arg-type]
    assert cfg.poll_interval == 3.0
    assert isinstance(cfg.poll_interval, float)


def test_slurm_config_poll_timeout_invalid_type():
    with pytest.raises(ConfigError, match="number"):
        SlurmConfig(poll_timeout=True)  # type: ignore[arg-type]


def test_slurm_config_poll_timeout_too_low():
    with pytest.raises(ConfigError, match="must be > 0"):
        SlurmConfig(poll_timeout=0)


def test_slurm_config_poll_timeout_int_coerced():
    cfg = SlurmConfig(poll_timeout=3600)  # type: ignore[arg-type]
    assert cfg.poll_timeout == 3600.0
    assert isinstance(cfg.poll_timeout, float)


def test_slurm_config_poll_retries_invalid_type():
    with pytest.raises(ConfigError, match="integer"):
        SlurmConfig(poll_retries=True)


def test_slurm_config_poll_retries_too_low():
    with pytest.raises(ConfigError, match="must be >= 1"):
        SlurmConfig(poll_retries=0)


def test_slurm_config_defaults_not_dict():
    with pytest.raises(ConfigError, match="table"):
        SlurmConfig(defaults="bad")  # type: ignore[arg-type]


def test_slurm_config_defaults_non_string_value():
    with pytest.raises(ConfigError, match="string"):
        SlurmConfig(defaults={"mem": 42})  # type: ignore[dict-item]



from __future__ import annotations

import argparse

from ..config import Config
from ..context import Context
from ..store.sqlite import SqliteBuildStore
from ..ui import Output
from .util import match_targets


def cmd_invalidate(
    args: argparse.Namespace, config: Config, ctx: Context, ui: Output
) -> int:
    targets = match_targets(ctx.tasks, args.pattern, config.default, args.regex)

    db_path = ctx.db_path
    if not db_path.exists():
        ui.status("No build store found, nothing to invalidate.")
        return 0

    with SqliteBuildStore(str(db_path)) as store:
        for task in targets:
            store.delete(task.task_id)
            ui.status(f"Invalidated [{task.name}]")

    return 0
